#!/usr/bin/python
# -*- utf-8 -*-

import parse
import math

class NetNode(object):
    def __init__(self, node):
        self.node = node.nid
        self.members = list()
        self.edges = list()

    def __repr__(self):
        return "NetNode(%s)" % str(self.node)

    def __str__(self):
        return str(self.node)

class NetEdge(object):

    def __init__(self, fromn, ton, way):
        self.from_node = fromn
        self.to_node = ton
        self.weight = 1.0
        self.oneway = False
        self.wayid = way.wid
        if 'name' in way.tag:
            self.name = way.tag['name']
        else:
            self.name = str(way.wid)
        self.oneway = self.check_onway(way)

    def __repr__(self):
        return "NetEdge(%s -- %s)" % (str(self.from_node), str(self.to_node))

    def calculate_weight(self, nodes, ways):
        way = ways[self.wayid]
        n1 = nodes[self.from_node.node]
        n2 = nodes[self.to_node.node]
        d = math.hypot(abs(n1.lat - n2.lat),
                       abs(n1.lon - n2.lon))
        # Set this to false to use classified weights
        if True:
            return d
        f = 1.0; maxspeed = 50
        if 'highway' in way.tag:
            highway = way.tag['highway']
            if highway == 'service':
                f *= 20; maxspeed = 30
            elif highway in ('steps'):
                f *= 10
            elif highway in ('footway', 'cycleway', 'track', 'pedestrian'):
                f *= 8; maxspeed = 20
            elif highway == 'residential':
                f *= 6; maxspeed = 35
            elif highway in ('unclassified', 'tertiary', 'teritary'):
                f *= 5; maxspeed = 50
            elif highway == 'secondary':
                f *= 4; maxspeed = 50
            elif highway == 'primary':
                f *= 3; maxspeed = 60
            elif highway in ('trunk', 'trunk_link'):
                f *= 1.5; maxspeed = 100
            elif highway in ('motorway', 'motorway_link'):
                f *= 1.0; maxspeed = 120
            else:
                raise Exception('Unknown highway=%s' % highway)
        else:
            f *= 1000
        if 'maxspeed' in way.tag:
            maxspeed = int(way.tag['maxspeed'])
        speedf = 1.0 + maxspeed * 0.25 / 120
        self.weight = f * speedf * d
#        print self.weight, f, speedf

    def check_onway(self, way):
        if 'oneway' in way.tag:
            if way.tag['oneway'] in ('yes', 'true', '1'):
                return True
            elif way.tag['oneway'] in ('no', 'false', '0'):
                return False

        if 'junction' in way.tag and way.tag['junction'] == 'roundabout':
            return True
        elif 'highway' in way.tag and way.tag['highway'] in ('motorway', 'motorway_link'):
            return True
        return False

def buildNet(parser, startnodeid):
    """
    This function creates a simplified network from parsed OSM file.
    The result is a network of nodes and edges. Where all nodes belonging
    to only one osm way are pool up into the new Nodes.
    """
    # we search the way the startnodes belongs to
    way = parser.nodes[startnodeid].member_of[0]

    seen = dict()
    seen[way] = way

    visit = list()

    net_nodes = list()
    net_edges = list()

    while True:
        lastn = None
        for node in way.nodes:
            if (node is way.nodes[0]) or (node is way.nodes[-1]):
                if node in seen:
                    n = seen[node]
                else:
                    n = NetNode(node)
                    n.members.append(way.wid)
                    net_nodes.append(n)
                    seen[node] = n
                if lastn is not None:
                    net_edges.append(NetEdge(lastn, n, way))
                lastn = n
            for member in node.member_of:
                if member is not way and isinstance(member, parse.Way):
                    if member not in seen:
                        if not member in visit:
                            visit.append(member) # XXX this cost too much performance
                    if node in seen:
                        n = seen[node]
                    else:
                        n = NetNode(node)
                        n.members.append(way.wid)
                        net_nodes.append(n)
                        seen[node] = n
                    if (lastn is not None) and not (lastn.node is n.node):
                        net_edges.append(NetEdge(lastn, n, way))
                    lastn = n
                        
        if len(visit) == 0:
            break
        way = visit.pop(0)
        seen[way] = way

    for edge in net_edges:
        # connect edges
        edge.from_node.edges.append(edge)
        if not edge.oneway:
            edge.to_node.edges.append(edge)
        # calculate edge weight
        edge.calculate_weight(parser.nodes, parser.ways)


    return (net_nodes, net_edges)

def buildPath(nodes, route):
    """
    This function get a path through the network and
    returns a new path that follows the affected roads
    """
    def addNodesInBetween(result, node, nextnode):
        n1 = nodes[node.node]
        n2 = nodes[nextnode.node]
        way = set(n1.member_of).intersection(n2.member_of).pop()
        idx1 = way.nodes.index(n1)
        idx2 = way.nodes.index(n2)
        if idx1 < idx2:
            for i in xrange(idx1, idx2):
                result.append(way.nodes[i])
        else:
            for i in xrange(idx1, idx2, -1):
                result.append(way.nodes[i])
       
    node = route.pop(0)
    nextnode = route.pop(0)
    result = list()
    addNodesInBetween(result, node, nextnode)
    while len(route) > 0:
        # step one further
        node = nextnode
        nextnode = route.pop(0)
        addNodesInBetween(result, node, nextnode) 
    result.append(nodes[nextnode.node])
    return result

