#!/usr/bin/python
# -*- utf-8 -*-

import xml.parsers.expat

class Node(object):
    def __init__(self, nid, lat, lon):
        self.nid = nid
        self.lat = lat
        self.lon = lon
        self.tag = dict()
        self.member_of = list()

    def __id__(self):
        return self.nid

    def __repr__(self):
        if 'name' in self.tag:
            return "Node(%ld, %s)" % (self.nid, self.name)
        else:
            return "Node(%ld)" % self.nid

    def __str__(self):
        return str(self.nid)

    def toxml(self):
        r = "<node id='%(nid)d' visible='true' lat='%(lat).13f' lon='%(lon).13f' " % vars(self)
        if len(self.tag) == 0:
            r += "/>\n"
        else:
            r += '>\n'
            for k, v in self.tag.iteritems():
                r += "  <tag k='%s' v='%s' />\n" % (k, v)
            r += "</node>\n"
        return r

class Way(object):
    def __init__(self, wid):
        self.wid = wid
        self.nodes = list()
        self.tag = dict()
        self.member_of = list()

    def __id__(self):
        return self.wid

    def __repr__(self):
        if 'name' in self.tag:
            return "Way(%ld, %s)" % (self.wid, self.tag['name'])
        else:
            return "Way(%ld)" % self.wid

    def __str__(self):
        return str(self.wid)

    def toxml(self):
        r = "<way id='%d' action='modify' visible='true'>\n" % self.wid
        for node in self.nodes:
            r += "  <nd ref='%d' />\n" % node.nid
        for k, v in self.tag.iteritems():
            r += "  <tag k='%s' v='%s' />\n" % (k, v)
        r += "</way>\n"
        return r


class Relation(object):
    def __init__(self, rid):
        self.rid = rid
        self.members = list()
        self.tag = dict()

    def __id__(self):
        return self.rid

    def __str__(self):
        return str(self.rid)

    def toxml(self):
        r = "<relation id='%(rid)d' visible='true'>\n"
        for member in self.members:
            if isinstance(member, Node):
                mt = 'node'
                ref = member.nid
            elif isinstance(member, Way):
                mt = 'way'
                ref = member.wid
            else:
                raise ValueError("Unknown member")
            r += "  <member type='%s' ref='%s' role='' />\n" % (mt, ref)
        for k, v in self.tag.iteritems():
            r += "  <tag k='%s' v='%s' />\n" % (k, v)
        r += "</relation>\n"
        return r


class OsmParser(object):
    def __init__(self):
        self.parser = xml.parsers.expat.ParserCreate()
        self.parser.StartElementHandler = self._start_element
        self.parser.EndElementHandler = self._end_element
        #self.parser.CharacterDataHandler = self._char_data

    def _start_element(self, name, attrs):
        #print 'Start element:', name, attrs
        if name == 'tag':
            k = attrs['k']
            v = attrs['v']
            if self.current_node is not None:
                self.current_node.tag[k] = v
            elif self.current_way is not None:
                self.current_way.tag[k] = v
            elif self.current_relation is not None:
                self.current_relation.tag[k] = v
            else:
                raise Exception('Parse error: tag is not attached to either a node, way or relation')
        elif name == 'node':
            nid = long(attrs['id'])
            self.current_node = Node(nid, float(attrs['lat']), float(attrs['lon']))
            self.nodes[nid] = self.current_node
        elif name == 'nd':
            if self.current_way is not None:
                nid = long(attrs['ref'])
                node = self.nodes[nid]
                node.member_of.append(self.current_way)
                self.current_way.nodes.append(node)
        elif name == 'way':
            wid = long(attrs['id'])
            self.current_way = Way(wid)
            self.ways[wid] = self.current_way
        elif name == 'member':
            if self.current_relation is not None:
                mid = long(attrs['ref'])
                if attrs['type'] == 'way':
                    member = self.ways[mid]
                elif attrs['type'] == 'node':
                    member = self.nodes[mid]
                else:
                    raise Exception('Unknown type for member')
                member.member_of.append(self.current_relation)
                self.current_relation.members.append(member)
        elif name == 'relation':
            rid = long(attrs['id'])
            self.current_relation = Relation(rid)
            self.relations[rid] = self.current_relation
        elif name == 'bound':
            pass
        elif name == 'osm':
            if float(attrs['version']) < 0.5:
                raise Exception("file version to old (< 0.5)")

    def _end_element(self, name):
        #print 'End element:', name
        if name == 'node':
            self.current_node = None
        elif name == 'way':
            self.current_way = None
        elif name == 'relation':
            self.current_relation = None
            
    def parseFile(self, fh):
        self.nodes = {}
        self.ways = {}
        self.relations = {}
        self.current_node = None
        self.current_way = None
        self.current_relation = None
        self.parser.ParseFile(fh)

# -------------------------------------------------------------
#  Functions to create OSM xml files
# -------------------------------------------------------------

def createXML(stream, nodes, ways, relations):
    stream.write("<?xml version='1.0' encoding='UTF-8'?>\n")
    stream.write("<osm version='0.5' generator='JOSM'>\n")
    for node in nodes:
        stream.write(node.toxml())
    for way in ways:
        stream.write(way.toxml())
    for rel in relations:
        stream.write(rel.toxml())
    stream.write("</osm>\n")

